import random
import argparse
import sys, string, glob, re, os
import numpy as np
import time,copy
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset,Dataset,DataLoader
from torch.nn import functional as F
from model_meta_gmm import MetaGMM

def test_epoch(xs,catess,model,args):
    model.eval()
    total_loss = 0.0
    total_ari = 0.0
    total_Kerr = 0.0
    cnt = 0
    for cates in catess:
        x = np.concatenate(xs[cates])
        y = np.empty((0))
        for i,cate in enumerate(cates):
            y = np.concatenate((y,np.ones(len(xs[cate]))*i))
        x = torch.FloatTensor(x).to(args.device)
        y = torch.LongTensor(y).to(args.device)
        y = onehot(y)
        loss,ari,x,r = model(x,y)
        thresh = 1e-1
        trueK = y.shape[0]
        estK = torch.unique(r.argmax(0)).shape[0]
        total_loss += loss.detach().item()
        total_ari += ari.detach().item()
        total_Kerr += np.abs(trueK-estK)
        cnt += 1
    return total_ari/cnt,total_Kerr/cnt

def onehot(y):
    n_sample = y.shape[0]
    n_class = y.max()+1
    vonehot = torch.LongTensor(n_class,n_sample).zero_().to(y.device)
    y = y.unsqueeze_(0)
    vonehot = vonehot.scatter_(0,y,1)
    return vonehot

def traineval_epoch(xs,cate,model,optimizer,args,is_train,is_proto=0):
    if is_train == 1:
        model.train()
    else:
        model.eval()

    total_loss = 0.0
    cnt = 0
    for batch_idx in range(int(len(xs)/(0.5*(args.minK+args.maxK)))):
        K = np.random.randint(args.minK,args.maxK)
        perm = np.random.permutation(len(cate))
        cates = cate[perm[:K]]
        x = np.concatenate(xs[cates])
        y = np.empty((0))
        for i,cate0 in enumerate(cates):
            y = np.concatenate((y,np.ones(len(xs[cate0]))*i))
        x = torch.FloatTensor(x).to(args.device)
        y = torch.LongTensor(y).to(args.device)
        y = onehot(y)
        if is_proto == 1:
            loss = model.proto(x,y)
        else:
            loss,ari,x,r = model(x,y)
        if is_train == 1 and args.use_encoder == 1:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        total_loss += loss.detach().item()
        cnt += 1
        
    return total_loss/cnt

def train(xs,train_cate,valid_cates,test_cates,model,optimizer,proto_optimizer,args):
    valid_losss = []

    for epoch_i in range(args.proto_epoch):
        train_loss = traineval_epoch(xs,train_cate,model,proto_optimizer,args,is_train=1,is_proto=1)
        valid_loss,valid_Kerr = test_epoch(xs,valid_cates,model,args)
        print("%d proto:train:loss: %f valid:loss: %f"%(epoch_i,train_loss,valid_loss))

        valid_losss += [valid_loss]
        model_state_dict = model.state_dict()
        checkpoint = {'model': model_state_dict,'settings': args,'epoch': epoch_i}
        if valid_loss >= max(valid_losss):
            bestmodel = copy.deepcopy(model)
            if args.out:
                torch.save(checkpoint,"%s.chkpt"%args.out)

    bestmodel = copy.deepcopy(model)
    model_state_dict = model.state_dict()
    checkpoint = {'model': model_state_dict,'settings': args,'epoch': 0}
    if args.out:
        torch.save(checkpoint,"%s.chkpt"%args.out)

    for epoch_i in range(args.epoch):
        train_loss = traineval_epoch(xs,train_cate,model,optimizer,args,is_train=1)
        valid_loss,valid_Kerr = test_epoch(xs,valid_cates,model,args)
        test_loss,valid_Kerr = test_epoch(xs,test_cates,model,args)

        print('%d train:loss: %f valid:loss: %f test:loss: %f'%(epoch_i,train_loss,valid_loss,test_loss))
                                                     
        valid_losss += [valid_loss]
        model_state_dict = model.state_dict()
        checkpoint = {
            'model': model_state_dict,
            'settings': args,
            'epoch': epoch_i}
        
        if valid_loss >= max(valid_losss):
            bestmodel = copy.deepcopy(model)
            if args.out:
                torch.save(checkpoint,"%s.chkpt"%args.out)

    return bestmodel


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--fn',type=str,default="",help='data file, e.g., data/csv/PenDigits_withoutdupl_norm_v01_t100/I10/fns0.npz')
    parser.add_argument('--epoch','-e',type=int,default=1000,help='epochs')
    parser.add_argument('--em_epoch',type=int,default=10,help='met-em epochs')
    parser.add_argument('--proto_epoch',type=int,default=1000,help='prototypical network epochs')
    parser.add_argument('--alpha0',type=float,default=1.0,help="Dirichlet parameter")
    parser.add_argument('--seed',type=int,default=0,help='random seed')
    parser.add_argument('--batch_size', type=int, default=128)
    parser.add_argument('--no_cuda', action='store_true')
    parser.add_argument('--lr', type=float, default=1e-5)
    parser.add_argument('--proto_lr', type=float, default=1e-2)    

    parser.add_argument('--weight_decay',type=float,default=1e-5)
    parser.add_argument('--dropout', type=float, default=0.1)
    parser.add_argument('--temperature', type=float, default=1)    
    parser.add_argument('--initmodel', default=None)
    parser.add_argument('--out', default=None)

    parser.add_argument('--use_encoder',type=int,default=1,help='use support set for neural net (1) or not (0)')
    parser.add_argument('--use_ari',type=int,default=1,help='use support set for neural net (1) or not (0)')
    parser.add_argument('--cal_r',type=int,default=0,help='calculate responsibility by neural net (1) or not (0)')
    
    parser.add_argument('--image',type=int,default=0,help='image(1) or not(0)')

    parser.add_argument('--n_dataset',type=int,default=0,help='number of tasks for training')
    parser.add_argument('--minK',type=int,default=2,help='minimum number of clusters')
    parser.add_argument('--maxK',type=int,default=10,help='minimum number of clusters')    
    parser.add_argument('--n_hiddens', type=str, default='256,2',
                        help="Comma separated sizes for hidden layers, e.g. 500, or 500,500")
    parser.add_argument('--hidden_size', type=int, default=32)    
                        
    args = parser.parse_args()    
    args.cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if not args.no_cuda and torch.cuda.is_available() else "cpu")
    args.device = device
    np.random.seed(args.seed)
    random.seed(args.seed)
    torch.manual_seed(args.seed)
    print(args)

    if re.search("omniglot",args.fn):
        args.image = 1
        
    heads = ['train','valid','test']
    dat = np.load(args.fn,allow_pickle=True)
    datad = {}
    datafn = str(dat['datafn'])
    xs = np.load(datafn,allow_pickle=True)['xs']
    train_cate = dat['train']
    valid_cates = dat['valids']
    test_cates = dat['tests']
    if args.n_dataset > 0:
        train_cate = train_cate[:args.n_dataset]

    n_att = xs[0].shape[1]
    n_class = len(train_cate)
        
    print("#train-cate: %d #valid: %d #test: %d #att: %d"%(len(train_cate),len(valid_cates),len(test_cates),n_att))

    args.n_hiddens = list(map(int,args.n_hiddens.split(',')))

    start = time.time()

    model = MetaGMM(n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,use_encoder=args.use_encoder,em_epoch=args.em_epoch,K=args.maxK,alpha0=args.alpha0,use_ari=args.use_ari,image=args.image,temperature=args.temperature,cal_r=args.cal_r,hidden_size=args.hidden_size,n_class=n_class)
    model.to(args.device)
    
    if args.initmodel:
        model.load_state_dict(torch.load("%s.chkpt"%args.initmodel)['model'],strict=False)        
        
    proto_optimizer = torch.optim.Adam(model.parameters(),lr=args.proto_lr,weight_decay=args.weight_decay)
    optimizer = torch.optim.Adam(model.parameters(),lr=args.lr,weight_decay=args.weight_decay)    

    model = train(xs,train_cate,valid_cates,test_cates,model,optimizer,proto_optimizer,args)

    train_time = (time.time()-start)
    start = time.time()
    test_loss,test_Kerr = test_epoch(xs,test_cates,model,args)    
    test_time = (time.time()-start)

    print("train:time: %f test:time: %f loss: %f Kerr: %f"%(train_time,test_time,test_loss,test_Kerr))
